"""
Random
====================================
Meta steps for hyperparameter tuning, such as random search.

..
   Copyright 2019, Neuraxio Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import copy
import math
from abc import ABC, abstractmethod
from typing import List, Callable, Tuple

import numpy as np
from sklearn.metrics import r2_score

from neuraxle.base import MetaStep, BaseStep, ExecutionContext, HandleOnlyMixin, ForceHandleOnlyMixin, \
    EvaluableStepMixin, _FittableStep
from neuraxle.data_container import DataContainer
from neuraxle.hyperparams.space import RecursiveDict
from neuraxle.steps.loop import StepClonerForEachDataInput
from neuraxle.steps.numpy import NumpyConcatenateOuterBatch, NumpyConcatenateOnCustomAxis

VALIDATION_SUB_DATA_CONTAINER_NAME = 'validation'


class BaseValidation(MetaStep, ABC):
    """
    Base class For validation wrappers.
    It has a scoring function to calculate the score for the validation split.

    .. seealso::
        :class`neuraxle.metaopt.random.ValidationSplitWrapper`,
        :class`Kneuraxle.metaopt.random.FoldCrossValidationWrapper`,
        :class`neuraxle.metaopt.random.AnchoredWalkForwardTimeSeriesCrossValidationWrapper`,
        :class`neuraxle.metaopt.random.WalkForwardTimeSeriesCrossValidationWrapper`

    """

    def __init__(self, wrapped=None, scoring_function: Callable = r2_score):
        """
        Base class For validation wrappers.
        It has a scoring function to calculate the score for the validation split.

        :param scoring_function: scoring function with two arguments (y_true, y_pred)
        :type scoring_function: Callable
        """
        BaseStep.__init__(self)
        MetaStep.__init__(self, wrapped)
        self.scoring_function = scoring_function

    @abstractmethod
    def split_data_container(self, data_container) -> Tuple[DataContainer, DataContainer]:
        pass


class BaseCrossValidationWrapper(EvaluableStepMixin, ForceHandleOnlyMixin, BaseValidation, ABC):
    # TODO: change default argument of scoring_function...
    def __init__(self, wrapped=None, scoring_function=r2_score, joiner=NumpyConcatenateOuterBatch(), cache_folder_when_no_handle=None,
                 split_data_container_during_fit=True, predict_after_fit=True):
        BaseValidation.__init__(self, wrapped=wrapped, scoring_function=scoring_function)
        ForceHandleOnlyMixin.__init__(self, cache_folder=cache_folder_when_no_handle)
        EvaluableStepMixin.__init__(self)

        self.split_data_container_during_fit = split_data_container_during_fit
        self.predict_after_fit = predict_after_fit
        self.joiner = joiner

    def train(self, train_data_container: DataContainer, context: ExecutionContext):
        step = StepClonerForEachDataInput(self.wrapped)
        step = step.handle_fit(train_data_container, context)

        return step

    def _fit_data_container(self, data_container: DataContainer, context: ExecutionContext) -> BaseStep:
        assert self.wrapped is not None

        step = StepClonerForEachDataInput(self.wrapped)
        step = step.handle_fit(data_container, context)

        return step

    def calculate_score(self, results):
        self.scores = [self.scoring_function(a, b) for a, b in zip(results.data_inputs, results.expected_outputs)]
        self.scores_mean = np.mean(self.scores)
        self.scores_std = np.std(self.scores)

    def split_data_container(self, data_container: DataContainer) -> Tuple[DataContainer, DataContainer]:
        train_data_inputs, train_expected_outputs, validation_data_inputs, validation_expected_outputs = self.split(
            data_container.data_inputs,
            data_container.expected_outputs
        )

        train_data_container = DataContainer(data_inputs=train_data_inputs, expected_outputs=train_expected_outputs)
        validation_data_container = DataContainer(
            data_inputs=validation_data_inputs,
            expected_outputs=validation_expected_outputs
        )

        return train_data_container, validation_data_container

    def get_score(self):
        return self.scores_mean

    def get_scores_std(self):
        return self.scores_std

    @abstractmethod
    def split(self, data_inputs, expected_outputs):
        raise NotImplementedError("TODO")


class ValidationSplitWrapper(BaseCrossValidationWrapper):
    """
    Wrapper for validation split that calculates the score for the validation split.

    .. code-block:: python

        random_search = Pipeline([
            RandomSearch(
                ValidationSplitWrapper(
                    Identity(),
                    test_size=0.1
                    scoring_function=mean_absolute_relative_error,
                    run_validation_split_in_test_mode=False
                ),
                n_iter= 10,
                higher_score_is_better= True,
                validation_technique=KFoldCrossValidationWrapper(),
                refit=True
            )
        ])

    .. note::
        The data is not shuffled before split. Please refer to the :class`DataShuffler` step for data shuffling.

    .. seealso::
        :class`BaseValidation`,
        :class`BaseCrossValidationWrapper`,
        :class`neuraxle.metaopt.auto_ml.RandomSearch`,
        :class`neuraxle.steps.data.DataShuffler`

    """

    def __init__(
            self,
            wrapped: BaseStep = None,
            test_size: float = 0.2,
            scoring_function=r2_score,
            run_validation_split_in_test_mode=True,
            cache_folder_when_no_handle=None
    ):
        """
        :param wrapped: wrapped step
        :param test_size: ratio for test size between 0 and 1
        :param scoring_function: scoring function with two arguments (y_true, y_pred)
        """
        BaseCrossValidationWrapper.__init__(self, wrapped=wrapped, cache_folder_when_no_handle=cache_folder_when_no_handle)

        self.run_validation_split_in_test_mode = run_validation_split_in_test_mode
        self.test_size = test_size
        self.scoring_function = scoring_function

    def _fit_data_container(self, data_container: DataContainer, context: ExecutionContext) -> ('ValidationSplitWrapper', DataContainer):
        """
        Fit using the training split.
        Calculate the scores using the validation split.

        :param context: execution context
        :param data_container: data container
        :type context: ExecutionContext
        :type data_container: DataContainer
        :return: fitted self
        """
        new_self, results_data_container = self._fit_transform_data_container(data_container, context)
        return new_self

    def _fit_transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> ('BaseStep', DataContainer):
        """
        Fit Transform given data inputs without splitting.

        :param context:
        :param data_container: DataContainer
        :type data_container: DataContainer
        :type context: ExecutionContext
        :return: outputs
        """
        train_data_container, validation_data_container = self.split_data_container(data_container)

        # add sub data container for the validation metrics calculated in MetricsWrapper
        train_data_container.add_sub_data_container(
            name=VALIDATION_SUB_DATA_CONTAINER_NAME,
            data_container=validation_data_container
        )

        self.wrapped, results_data_container = self.wrapped.handle_fit_transform(train_data_container,
                                                                                 context.push(self.wrapped))

        self._update_scores_train(results_data_container.data_inputs, results_data_container.expected_outputs)

        results_data_container = self.wrapped.handle_predict(validation_data_container, context.push(self.wrapped))

        self._update_scores_validation(results_data_container.data_inputs, results_data_container.expected_outputs)

        self.wrapped.apply('disable_metrics')
        data_container = self.wrapped.handle_predict(data_container, context.push(self.wrapped))
        self.wrapped.apply('enable_metrics')

        return self, data_container

    def _transform_data_container(self, data_container: DataContainer, context: ExecutionContext):
        """
        Transform given data inputs without splitting.

        :param context: execution context
        :param data_container: DataContainer
        :type data_container: DataContainer
        :type context: ExecutionContext
        :return: outputs
        """
        return self.wrapped.handle_transform(data_container, context.push(self.wrapped))

    def _update_scores_validation(self, data_inputs, expected_outputs):
        self.scores_validation = self.scoring_function(expected_outputs, data_inputs)
        self.scores_validation_mean = np.mean(self.scores_validation)
        self.scores_validation_std = np.std(self.scores_validation)

    def _update_scores_train(self, data_inputs, expected_outputs):
        self.scores_train = self.scoring_function(expected_outputs, data_inputs)
        self.scores_train_mean = np.mean(self.scores_train)
        self.scores_train_std = np.std(self.scores_train)

    def get_score(self):
        return self.scores_validation_mean

    def get_score_validation(self):
        return self.scores_validation_mean

    def get_score_train(self):
        return self.scores_validation_mean

    def split_data_container(self, data_container) -> Tuple[DataContainer, DataContainer]:
        """
        Split data container into a training set, and a validation set.

        :param data_container: data container
        :type data_container: DataContainer
        :return: train_data_container, validation_data_container
        """

        train_data_inputs, train_expected_outputs, validation_data_inputs, validation_expected_outputs = \
            self.split(data_container.data_inputs, data_container.expected_outputs)

        train_ids = self.train_split(data_container.current_ids)
        train_data_container = DataContainer(
            data_inputs=train_data_inputs,
            current_ids=train_ids,
            summary_id=data_container.summary_id,
            expected_outputs=train_expected_outputs
        )

        validation_ids = self.validation_split(data_container.current_ids)
        validation_data_container = DataContainer(
            data_inputs=validation_data_inputs,
            current_ids=validation_ids,
            summary_id=data_container.summary_id,
            expected_outputs=validation_expected_outputs
        )

        return train_data_container, validation_data_container

    def split(self, data_inputs, expected_outputs=None) -> Tuple[List, List, List, List]:
        """
        Split data inputs, and expected outputs into a training set, and a validation set.

        :param data_inputs: data inputs to split
        :param expected_outputs: expected outputs to split
        :return: train_data_inputs, train_expected_outputs, validation_data_inputs, validation_expected_outputs
        """
        validation_data_inputs = self.validation_split(data_inputs)
        validation_expected_outputs = None
        if expected_outputs is not None:
            validation_expected_outputs = self.validation_split(expected_outputs)

        train_data_inputs = self.train_split(data_inputs)
        train_expected_outputs = None
        if expected_outputs is not None:
            train_expected_outputs = self.train_split(expected_outputs)

        return train_data_inputs, train_expected_outputs, validation_data_inputs, validation_expected_outputs

    def train_split(self, data_inputs) -> List:
        """
        Split training set.

        :param data_inputs: data inputs to split
        :return: train_data_inputs
        """
        return data_inputs[0:self._get_index_split(data_inputs)]

    def validation_split(self, data_inputs) -> List:
        """
        Split validation set.

        :param data_inputs: data inputs to split
        :return: validation_data_inputs
        """
        return data_inputs[self._get_index_split(data_inputs):]

    def disable_metrics(self):
        self.metrics_enabled = False
        if self.wrapped is not None:
            self.wrapped.apply('disable_metrics')
        return RecursiveDict()

    def enable_metrics(self):
        self.metrics_enabled = True
        if self.wrapped is not None:
            self.wrapped.apply('enable_metrics')
        return RecursiveDict()

    def _get_index_split(self, data_inputs):
        return math.floor(len(data_inputs) * (1 - self.test_size))


def average_kfold_scores(metric_function):
    def calculate(y_true_kfolds, y_pred_kfolds):
        kfold_scores = []
        for y_true, y_pred in zip(y_true_kfolds, y_pred_kfolds):
            kfold_scores.append(metric_function(y_true, y_pred))

        return np.mean(kfold_scores)

    return calculate


class KFoldCrossValidationWrapper(BaseCrossValidationWrapper):
    def __init__(
            self,
            scoring_function=r2_score,
            k_fold=3,
            joiner=NumpyConcatenateOuterBatch(),
            cache_folder_when_no_handle=None
    ):
        self.k_fold = k_fold
        BaseCrossValidationWrapper.__init__(
            self,
            scoring_function=scoring_function,
            joiner=joiner,
            cache_folder_when_no_handle=cache_folder_when_no_handle
        )

    def split(self, data_inputs, expected_outputs):
        validation_data_inputs, validation_expected_outputs = self.validation_split(data_inputs, expected_outputs)
        train_data_inputs, train_expected_outputs = self.train_split(data_inputs, expected_outputs)

        return train_data_inputs, train_expected_outputs, validation_data_inputs, validation_expected_outputs

    def train_split(self, data_inputs, expected_outputs) -> (List, List):
        train_data_inputs = []
        train_expected_outputs = []
        data_inputs = np.array(data_inputs)
        expected_outputs = np.array(expected_outputs)

        for i in range(len(data_inputs)):
            before_di = data_inputs[:i]
            after_di = data_inputs[i + 1:]
            inputs = (before_di, after_di)

            before_eo = expected_outputs[:i]
            after_eo = expected_outputs[i + 1:]
            outputs = (before_eo, after_eo)

            inputs = self.joiner.transform(inputs)
            outputs = self.joiner.transform(outputs)

            train_data_inputs.append(inputs)
            train_expected_outputs.append(outputs)

        return train_data_inputs, train_expected_outputs

    def validation_split(self, data_inputs, expected_outputs=None) -> (List, List):
        splitted_data_inputs = self._split(data_inputs)
        if expected_outputs is not None:
            splitted_expected_outputs = self._split(expected_outputs)
            return splitted_data_inputs, splitted_expected_outputs

        return splitted_data_inputs, [None] * len(splitted_data_inputs)

    def _split(self, data_inputs):
        splitted_data_inputs = []
        step = len(data_inputs) / float(self.k_fold)
        for i in range(self.k_fold):
            a = int(step * i)
            b = int(step * (i + 1))
            if b > len(data_inputs):
                b = len(data_inputs)

            slice = data_inputs[a:b]
            splitted_data_inputs.append(slice)
        return splitted_data_inputs


class AnchoredWalkForwardTimeSeriesCrossValidationWrapper(BaseCrossValidationWrapper):
    """
    Prform an anchored walk forward cross validation by performing a forward rolling split.
    All training splits start at the beginning of the time series, but finish at different time. The finish time
    increase toward the end at each forward split.

    For the validation split it will start after a certain time delay (if padding is set)
    after their corresponding training split.

    Notes: The data supported by this cross validation is nd.array of shape [batch_size, total_time_steps, n_features].
    The array can have an arbitrary number of dimension, but the time series axis is currently limited to `axis=1`.
    """

    def __init__(self, minimum_training_size, validation_window_size=None, padding_between_training_and_validation=0,
                 drop_remainder=False, scoring_function=r2_score, joiner=NumpyConcatenateOnCustomAxis(axis=1)):
        """
        Create a anchored walk forward time series cross validation object.

        The size of the validation split is defined by `validation_window_size`.
        The difference in start position between two consecutive validation split is also equal to
        `validation_window_size`.

        :param minimum_training_size: size of the smallest training split.
        :param validation_window_size: size of each validation split and also the time step taken between each
            forward roll, by default None. If None : It takes the value `minimum_training_size`.
        :param padding_between_training_and_validation: the size of the padding between the end of the training split
            and the start of the validation split, by default 0.
        :param drop_remainder: drop the last split if the last validation split does not coincide
            with a full validation_window_size, by default False.
        :param scoring_function: scoring function use to validate performance if it is not None, by default r2_score,
        :param joiner the joiner callable that can join the different result together.
        :return: WalkForwardTimeSeriesCrossValidation instance.
        """
        BaseCrossValidationWrapper.__init__(self, scoring_function=scoring_function, joiner=joiner)
        self.minimum_training_size = minimum_training_size
        # If validation_window_size is None, we give the same value as training_window_size.
        self.validation_window_size = validation_window_size or self.minimum_training_size
        self.padding_between_training_and_validation = padding_between_training_and_validation
        self.drop_remainder = drop_remainder
        self._validation_initial_start = self.minimum_training_size + self.padding_between_training_and_validation

    def split(self, data_inputs, expected_outputs):
        """
        Split the data into train inputs, train expected outputs, validation inputs, validation expected outputs.

        Notes: The data supported by this cross validation is nd.array of shape [batch_size, total_time_steps, n_features].
        The array can have an arbitrary number of dimension, but the time series axis is currently limited to `axis=1`.

        :param data_inputs: data to perform walk forward cross validation into.
        :param expected_outputs: the expected target/label that will be used during walk forward cross validation.
        :return: train_data_inputs, train_expected_outputs, validation_data_inputs, validation_expected_outputs
        """
        validation_data_inputs, validation_expected_outputs = self.validation_split(
            data_inputs, expected_outputs)

        train_data_inputs, train_expected_outputs = self.train_split(
            data_inputs, expected_outputs)

        return train_data_inputs, train_expected_outputs, validation_data_inputs, validation_expected_outputs

    def train_split(self, data_inputs, expected_outputs=None) -> (List, List):
        """
        Split the data into train inputs, train expected outputs

        Notes: The data supported by this cross validation is nd.array of shape [batch_size, total_time_steps, n_features].
        The array can have an arbitrary number of dimension, but the time series axis is currently limited to `axis=1`.

        :param data_inputs: data to perform walk forward cross validation into.
        :param expected_outputs: the expected target*label that will be used during walk forward cross validation.
        :return: train_data_inputs, train_expected_outputs
        """
        splitted_data_inputs = self._train_split(data_inputs)
        if expected_outputs is not None:
            splitted_expected_outputs = self._train_split(expected_outputs)
            return splitted_data_inputs, splitted_expected_outputs

    def validation_split(self, data_inputs, expected_outputs=None) -> List:
        """
        Split the data into validation inputs, validation expected outputs.

        Notes: The data supported by this cross validation is nd.array of shape [batch_size, total_time_steps, n_features].
        The array can have an arbitrary number of dimension, but the time series axis is currently limited to `axis=1`.

        :param data_inputs: data to perform walk forward cross validation into.
        :param expected_outputs: the expected target*label that will be used during walk forward cross validation.
        :return: validation_data_inputs, validation_expected_outputs
        """
        splitted_data_inputs = self._validation_split(data_inputs)
        if expected_outputs is not None:
            splitted_expected_outputs = self._validation_split(expected_outputs)
            return splitted_data_inputs, splitted_expected_outputs
        return splitted_data_inputs

    def _train_split(self, data_inputs):
        splitted_data_inputs = []
        number_step = self._get_number_fold(data_inputs)

        for i in range(number_step):
            # first slice index is always 0 for anchored walk forward cross validation.
            a = 0
            b = int(self.minimum_training_size + i * self.validation_window_size)

            if b > data_inputs.shape[1]:
                b = data_inputs.shape[1]

            # TODO: The slicer could a inverse_transform of the joiner. A len method should also be defined.
            slice = data_inputs[:, a:b]
            splitted_data_inputs.append(slice)
        return splitted_data_inputs

    def _validation_split(self, data_inputs):
        splitted_data_inputs = []
        number_step = self._get_number_fold(data_inputs)
        for i in range(number_step):
            a = int(self._validation_initial_start + i * self.validation_window_size)
            b = int(a + self.validation_window_size)

            if b > data_inputs.shape[1]:
                b = data_inputs.shape[1]

            # TODO: The slicer could a inverse_transform of the joiner. A len method should also be defined.
            slice = data_inputs[:, a:b]
            splitted_data_inputs.append(slice)
        return splitted_data_inputs

    def _get_number_fold(self, data_inputs):
        if self.drop_remainder:
            number_step = math.floor(
                (data_inputs.shape[1] - self._validation_initial_start) / float(self.validation_window_size)
            )
        else:
            number_step = math.ceil(
                (data_inputs.shape[1] - self._validation_initial_start) / float(self.validation_window_size)
            )
        return number_step


class WalkForwardTimeSeriesCrossValidationWrapper(AnchoredWalkForwardTimeSeriesCrossValidationWrapper):
    """
    Perform a classic walk forward cross validation by performing a forward rolling split.

    All the training split have the same `validation_window_size` size. The start time and end time of each training
    split will increase identically toward the end at each forward split. Same principle apply with the validation
    split, where the start and end will increase in the same manner toward the end. Each validation split start after
    a certain time delay (if padding is set) after their corresponding training split.

    Notes: The data supported by this cross validation is nd.array of shape [batch_size, total_time_steps, n_features].
    The array can have an arbitrary number of dimension, but the time series axis is currently limited to `axis=1`.
    """

    def __init__(self, training_window_size, validation_window_size=None, padding_between_training_and_validation=0,
                 drop_remainder=False, scoring_function=r2_score, joiner=NumpyConcatenateOnCustomAxis(axis=1)):
        """
        Create a classic walk forward time series cross validation object.

        The difference in start position between two consecutive validation split are equal to one
        `validation_window_size`.

        :param training_window_size: the window size of training split.
        :param validation_window_size: the window size of each validation split and also the time step taken between
            each forward roll, by default None. If None : It takes the value `training_window_size`.
        :param padding_between_training_and_validation: the size of the padding between the end of the training split
            and the start of the validation split, by default 0.
        :param drop_remainder: drop the last split if the last validation split does not coincide
            with a full validation_window_size, by default False.
        :param scoring_function: scoring function use to validate performance if it is not None, by default r2_score,
        :param joiner the joiner callable that can join the different result together.
        :return: WalkForwardTimeSeriesCrossValidation instance.
        """
        AnchoredWalkForwardTimeSeriesCrossValidationWrapper.__init__(
            self,
            training_window_size,
            validation_window_size=validation_window_size,
            padding_between_training_and_validation=padding_between_training_and_validation,
            drop_remainder=drop_remainder, scoring_function=scoring_function, joiner=joiner
        )

    def _train_split(self, data_inputs):
        splitted_data_inputs = []
        number_step = self._get_number_fold(data_inputs)

        for i in range(number_step):
            a = int(i * self.validation_window_size)
            # Here minimum_training_size = training_size, since each training split has the same length.
            b = int(a + self.minimum_training_size)

            if b > data_inputs.shape[1]:
                b = data_inputs.shape[1]

            # TODO: The slicer could a inverse_transform of the joiner. A len method should also be defined.
            slice = data_inputs[:, a:b]
            splitted_data_inputs.append(slice)
        return splitted_data_inputs



"""
Hyperparameter Dictionary Conversions
=====================================
Ways to convert from a nested dictionary of hyperparameters to a flat dictionary, and vice versa.

Here is a nested dictionary:

.. code-block:: python

    {
        "b": {
            "a": {
                "learning_rate": 7
            },
            "learning_rate": 9
        }
    }

Here is an equivalent flat dictionary for the previous nested one:

.. code-block:: python

    {
        "b.a.learning_rate": 7,
        "b.learning_rate": 9
    }

Notice that if you have a ``SKLearnWrapper`` on a sklearn Pipeline object, the hyperparameters past that point will use
double underscores ``__`` as a separator rather than a dot in flat dictionaries, and in nested dictionaries the
sklearn params will appear as a flat past the sklearn wrapper, which is fine.

By default, hyperparameters are stored inside a HyperparameterSpace or inside a HyperparameterSamples object, which
offers methods to do the conversions above, and also using ordered dicts (OrderedDict) to store parameters in-order.

A HyperparameterSpace can be sampled by calling the ``.rvs()`` method on it, which will recursively call ``.rvs()`` on all
the HyperparameterSpace and HyperparameterDistribution that it contains. It will return a HyperparameterSamples object.
A HyperparameterSpace can also be narrowed towards an better, finer subspace, which is itself a HyperparameterSpace.
This can be done by calling the ``.narrow_space_from_best_guess`` method of the HyperparameterSpace which will also
recursively apply the changes to its contained HyperparameterSpace and to all its contained HyperparameterDistribution.

The HyperparameterSamples contains sampled hyperparameter, that is, a valued point in the possible space. This is
ready to be sent to an instance of the pipeline to try and score it, for example.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

from collections import OrderedDict

from scipy.stats import rv_continuous, rv_discrete
from scipy.stats._distn_infrastructure import rv_generic

from neuraxle.hyperparams.distributions import HyperparameterDistribution
from neuraxle.hyperparams.scipy_distributions import ScipyDiscreteDistributionWrapper, \
    ScipyContinuousDistributionWrapper


class RecursiveDict(OrderedDict):
    """
    Wraps an hyperparameter nested dict or flat dict, and offer a few more functions.

    This can be set on a Pipeline with the method ``set_hyperparams``.

    HyperparameterSamples are often the result of calling ``.rvs()`` on an HyperparameterSpace.
    """
    DEFAULT_SEPARATOR = '__'

    def __init__(self, *args, separator=None, **kwds):
        args = self._patch_args(*args)

        if len(args) == 1 and isinstance(args[0], RecursiveDict) and len(kwds) == 0:
            super().__init__(args[0].items())
            separator = args[0].separator
        else:
            super().__init__(*args, **kwds)

        if separator is None:
            separator = self.DEFAULT_SEPARATOR

        self.separator = separator

    def _patch_args(self, *args):
        patched_args = []
        for arg in args:
            if isinstance(arg, RecursiveDict):
                patched_arg = RecursiveDict(**{
                    name: self._patch_arg(value)
                    for name, value in args[0].items()
                }, separator=arg.separator)
            else:
                patched_arg = arg
            patched_args.append(patched_arg)

        return patched_args

    def _patch_arg(self, arg):
        if hasattr(arg, 'dist') and isinstance(arg.dist, rv_generic):
            if isinstance(arg.dist, rv_discrete):
                return ScipyDiscreteDistributionWrapper(arg)
            if isinstance(arg.dist, rv_continuous):
                return ScipyContinuousDistributionWrapper(arg)
        else:
            return arg

    def __getitem__(self, item: str = None):
        item_values = type(self)()

        for name, values in self.items():
            if item is None and not self.separator in name:
                item_values[name] = values

            if self.separator in name:
                name_split = name.split(self.separator)
                if str(name_split[0]) == item:
                    item_values[self.separator.join(name_split[1:])] = values

            if item == name:
                item_values = values

        return item_values

    def to_flat(self) -> 'RecursiveDict':
        """
        Will create an equivalent flat HyperparameterSamples.

        :return: an HyperparameterSamples like self, flattened.
        """
        return self.nested_dict_to_flat(dict_ctor=type(self))

    def to_nested_dict(self) -> 'RecursiveDict':
        """
        Will create an equivalent nested dict HyperparameterSamples.

        :return: an HyperparameterSamples like self, as a nested dict.
        """
        return self.flat_to_nested_dict(dict_ctor=type(self))

    def to_flat_as_dict_primitive(self) -> dict:
        """
        Will create an equivalent flat HyperparameterSpace, as a dict.

        :return: an HyperparameterSpace like self, flattened.
        """
        return self.nested_dict_to_flat(dict_ctor=dict)

    def to_nested_dict_as_dict_primitive(self) -> dict:
        """
        Will create an equivalent nested dict HyperparameterSpace, as a dict.

        :return: a nested primitive dict type of self.
        """
        return self.flat_to_nested_dict(dict_ctor=dict)

    def to_flat_as_ordered_dict_primitive(self) -> OrderedDict:
        """
        Will create an equivalent flat HyperparameterSpace, as a dict.

        :return: an HyperparameterSpace like self, flattened.
        """
        return self.nested_dict_to_flat(dict_ctor=OrderedDict)

    def to_nested_dict_as_ordered_dict_primitive(self) -> OrderedDict:
        """
        Will create an equivalent nested dict HyperparameterSpace, as a dict.

        :return: a nested primitive dict type of self.
        """
        return self.flat_to_nested_dict(dict_ctor=OrderedDict)

    def nested_dict_to_flat(self, dict_ctor):
        """
        Convert a nested hyperparameter dictionary to a flat one.

        :param dict_ctor: ``OrderedDict`` by default. Will use this as a class to create the new returned dict.
        :return: a flat hyperparameter dictionary.
        """
        if isinstance(dict_ctor, RecursiveDict):
            ret = dict_ctor(self.separator)
        else:
            ret = dict_ctor()

        for k, v in self.items():
            if isinstance(v, dict) or isinstance(v, OrderedDict) or isinstance(v, dict_ctor):
                nested_dict = v
                if not isinstance(v, RecursiveDict):
                    nested_dict = RecursiveDict(v, separator=self.separator)
                _ret = nested_dict.nested_dict_to_flat(dict_ctor=type(nested_dict))
                for key, val in _ret.items():
                    ret[k + nested_dict.separator + key] = val
            else:
                ret[k] = v
        return ret

    def flat_to_nested_dict(self, dict_ctor=dict):
        """
        Convert a flat hyperparameter dictionary to a nested one.

        :param dict_ctor: ``OrderedDict`` by default. Will use this as a class to create the new returned dict.
        :return: a nested hyperparameter dictionary.
        """
        if isinstance(dict_ctor, RecursiveDict):
            pre_ret = dict_ctor(self.separator)
            ret = dict_ctor(self.separator)
        else:
            pre_ret = dict_ctor()
            ret = dict_ctor()

        for k, v in self.items():
            k, _, key = k.partition(self.separator)
            if len(key) > 0:
                if k not in pre_ret.keys():
                    pre_ret[k] = dict_ctor()
                pre_ret[k][key] = v
            else:
                ret[k] = v
        for k, v in pre_ret.items():
            flat_dict = v
            if not isinstance(v, RecursiveDict):
                flat_dict = RecursiveDict(v, separator=self.separator)
            ret[k] = flat_dict.flat_to_nested_dict(dict_ctor=dict_ctor)
        return ret

    def with_separator(self, separator):
        """
        Create a new recursive dict that uses the given separator at each level.

        :param separator:
        :return:
        """
        return type(self)(
            separator=separator,
            **{
                key.replace(self.separator, separator): value if not isinstance(value, RecursiveDict) \
                    else value.with_separator(separator) for key, value in self.items()
            })


class HyperparameterSamples(RecursiveDict):
    """
    Wraps an hyperparameter nested dict or flat dict, and offer a few more functions.

    This can be set on a Pipeline with the method ``set_hyperparams``.

    HyperparameterSamples are often the result of calling ``.rvs()`` on an HyperparameterSpace.
    """

    def __init__(self, *args, separator=None, **kwds):
        super().__init__(*args, separator=separator, **kwds)


class HyperparameterSpace(RecursiveDict):
    """
    Wraps an hyperparameter nested dict or flat dict, and offer a few more functions to process
    all contained HyperparameterDistribution.

    This can be set on a Pipeline with the method ``set_hyperparams_space``.

    Calling ``.rvs()`` on an ``HyperparameterSpace`` results in ``HyperparameterSamples``.
    """

    def __init__(self, *args, separator=None, **kwds):
        super().__init__(*args, separator=separator, **kwds)

    def rvs(self) -> 'HyperparameterSamples':
        """
        Sample the space of random variables.

        :return: a random HyperparameterSamples, sampled from a point of the present HyperparameterSpace.
        """
        new_items = []
        for k, v in self.items():
            if isinstance(v, HyperparameterDistribution) or isinstance(v, HyperparameterSpace):
                v = v.rvs()
            new_items.append((k, v))
        return HyperparameterSamples(new_items)

    def nullify(self):
        new_items = []
        for k, v in self.items():
            if isinstance(v, HyperparameterDistribution) or isinstance(v, HyperparameterSpace):
                v = v.nullify()
            new_items.append((k, v))
        return HyperparameterSamples(new_items)

    def narrow_space_from_best_guess(
        self,
        best_guesses: 'HyperparameterSpace',
        kept_space_ratio: float = 0.5
    ) -> 'HyperparameterSpace':
        """
        Takes samples estimated to be the best ones of the space as of yet, and restrict the whole space towards that.

        :param best_guesses: sampled HyperparameterSpace (the result of rvs on each parameter, but still stored as a HyperparameterSpace).
        :param kept_space_ratio: what proportion of the space is kept. Should be between 0.0 and 1.0. Default is 0.5.
        :return: a new HyperparameterSpace containing the narrowed HyperparameterDistribution objects.
        """
        new_items = []
        for k, v in self.items():
            if isinstance(v, HyperparameterDistribution) or isinstance(v, HyperparameterSpace):
                best_guess_v = best_guesses[k]
                v = v.narrow_space_from_best_guess(best_guess_v, kept_space_ratio)
            new_items.append((k, v))
        return HyperparameterSpace(new_items)

    def unnarrow(self) -> 'HyperparameterSpace':
        """
        Return the original space before narrowing of the distribution. If the distribution was never narrowed,
        the values in the dict will be copies.

        :return: the original HyperparameterSpace before narrowing.
        """
        new_items = []
        for k, v in self.items():
            if isinstance(v, HyperparameterDistribution) or isinstance(v, HyperparameterSpace):
                v = v.unnarrow()
            new_items.append((k, v))
        return HyperparameterSpace(new_items)

    def to_flat(self) -> 'HyperparameterSpace':
        """
        Will create an equivalent flat HyperparameterSpace.

        :return: an HyperparameterSpace like self, flattened.
        """
        return self.nested_dict_to_flat(dict_ctor=HyperparameterSpace)

    def to_nested_dict(self) -> 'HyperparameterSpace':
        """
        Will create an equivalent nested dict HyperparameterSpace.

        :return: an HyperparameterSpace like self, as a nested dict.
        """
        return self.flat_to_nested_dict(dict_ctor=HyperparameterSpace)



"""
Scikit-learn metaoptimizers
=====================================
E.g.: for use with RandomizedSearchCV.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

from neuraxle.base import MetaStep, BaseStep


class MetaSKLearnWrapper(MetaStep):

    def __init__(self, wrapped):
        """
        Wrap a scikit-learn MetaEstimatorMixin for usage in Neuraxle. 
        This class is similar to the SKLearnWrapper class of Neuraxle that can wrap a scikit-learn BaseEstimator. 
        
        :param wrapped: a scikit-learn object of type "MetaEstimatorMixin". 
        """
        MetaStep.__init__(self)

        self.wrapped_sklearn_metaestimator = wrapped  # TODO: use self.set_step of the MetaStepMixin instead?
        # sklearn.model_selection.RandomizedSearchCV

    def fit(self, data_inputs, expected_outputs=None) -> 'BaseStep':
        self.wrapped_sklearn_metaestimator = self.wrapped_sklearn_metaestimator.fit(data_inputs, expected_outputs)
        return self

    def transform(self, data_inputs):
        return self.wrapped_sklearn_metaestimator.transform(data_inputs)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        type_ = self.__class__
        module = type_.__module__
        qualname = type_.__qualname__
        wrappedname = str(self.wrapped_sklearn_metaestimator.__class__.__name__)
        return "<{}.{}({}(...)) object {}>".format(module, qualname, wrappedname, hex(id(self)))



"""
Neuraxle's DataContainer classes
====================================
Classes for containing the data that flows throught the pipeline steps.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import hashlib
import math
from typing import Any, Iterable, List, Tuple, Union

import numpy as np
from conv import convolved_1d

NamedDataContainerTuple = Tuple[str, 'DataContainer']


class DataContainer:
    """
    DataContainer class to store data inputs, expected outputs, and ids together.
    Each :class:`BaseStep` needs to rehash ids with hyperparameters so that the :class:`Checkpoint` step
    can create checkpoints for a set of hyperparameters.

    The DataContainer object is passed to all of the :class:`BaseStep` handle methods :
        * :func:`~neuraxle.base.BaseStep.handle_transform`
        * :func:`~neuraxle.base.BaseStep.handle_fit_transform`
        * :func:`~neuraxle.base.BaseStep.handle_fit`

    Most of the time, you won't need to care about the DataContainer because it is the pipeline that manages it.

    .. seealso::
        :class:`~neuraxle.base.BaseHasher`,
        :class: `neuraxle.base.BaseStep`
    """

    def __init__(
            self,
            data_inputs: Any,
            current_ids=None,
            summary_id=None,
            expected_outputs: Any = None,
            sub_data_containers: List['NamedDataContainerTuple'] = None
    ):
        self.summary_id = summary_id
        self.data_inputs = data_inputs

        if current_ids is None:
            if hasattr(data_inputs, '__len__'):
                current_ids = [str(c) for c in range(len(data_inputs))]
            else:
                current_ids = str(0)

        self.current_ids = current_ids

        if expected_outputs is None and isinstance(data_inputs, Iterable):
            self.expected_outputs = [None] * len(data_inputs)
        else:
            self.expected_outputs = expected_outputs

        if sub_data_containers is None:
            sub_data_containers = []

        self.sub_data_containers: List[NamedDataContainerTuple] = sub_data_containers

    def set_data_inputs(self, data_inputs: Iterable):
        """
        Set data inputs.

        :param data_inputs: data inputs
        :type data_inputs: Iterable
        :return:
        """
        self.data_inputs = data_inputs
        return self

    def set_expected_outputs(self, expected_outputs: Iterable):
        """
        Set expected outputs.

        :param expected_outputs: expected outputs
        :type expected_outputs: Iterable
        :return:
        """
        self.expected_outputs = expected_outputs
        return self

    def set_current_ids(self, current_ids: List[str]):
        """
        Set current ids.

        :param current_ids: data inputs
        :type current_ids: List[str]
        :return:
        """
        self.current_ids = current_ids
        return self

    def set_summary_id(self, summary_id: str):
        """
        Set summary id.

        :param summary_id: str
        :return:
        """
        self.summary_id = summary_id
        return self

    def set_sub_data_containers(self, sub_data_containers: List['DataContainer']):
        """
        Set sub data containers
        :return:
        """
        self.sub_data_containers = sub_data_containers

    def hash_summary(self):
        """
        Hash :class:`DataContainer`.current_ids, data inputs, and hyperparameters together into one id.

        :return: single hashed current id for all of the current ids
        :rtype: str
        """
        m = hashlib.md5()
        for current_id in self.current_ids:
            m.update(str.encode(str(current_id)))
        return m.hexdigest()

    def convolved_1d(self, stride, kernel_size) -> Iterable['DataContainer']:
        """
        Returns an iterator that iterates through batches of the DataContainer.

        :param stride: step size for the convolution operation
        :param kernel_size:
        :return: an iterator of DataContainer
        :rtype: Iterable[DataContainer]

        .. seealso::
            `<https://github.com/guillaume-chevalier/python-conv-lib>`_
        """
        conv_current_ids = convolved_1d(stride=stride, iterable=self.current_ids, kernel_size=kernel_size,
                                        include_incomplete_pass=True)
        conv_data_inputs = convolved_1d(stride=stride, iterable=self.data_inputs, kernel_size=kernel_size,
                                        include_incomplete_pass=True)
        conv_expected_outputs = convolved_1d(stride=stride, iterable=self.expected_outputs, kernel_size=kernel_size,
                                             include_incomplete_pass=True)

        for current_ids, data_inputs, expected_outputs in zip(conv_current_ids, conv_data_inputs,
                                                              conv_expected_outputs):
            for i, (ci, di, eo) in enumerate(zip(current_ids, data_inputs, expected_outputs)):
                if di is None:
                    current_ids = current_ids[:i]
                    data_inputs = data_inputs[:i]
                    expected_outputs = expected_outputs[:i]
                    break

            yield DataContainer(
                data_inputs=data_inputs,
                current_ids=current_ids,
                summary_id=self.summary_id,
                expected_outputs=expected_outputs,
                sub_data_containers=self.sub_data_containers
            )

    def get_n_batches(self, batch_size: int) -> int:
        return math.ceil(len(self.data_inputs) / batch_size)

    def copy(self):
        return DataContainer(
            data_inputs=self.data_inputs,
            current_ids=self.current_ids,
            summary_id=self.summary_id,
            expected_outputs=self.expected_outputs,
            sub_data_containers=[(name, data_container.copy()) for name, data_container in self.sub_data_containers]
        )

    def add_sub_data_container(self, name: str, data_container: 'DataContainer'):
        """
        Get sub data container if item is str, otherwise get a zip of current ids, data inputs, and expected outputs.

        :type name: sub data container name
        :type data_container: sub data container
        :return: self
        """
        self.sub_data_containers.append((name, data_container))
        return self

    def get_sub_data_container_names(self):
        """
        Get sub data container names.

        :return: list of names
        """
        return [name for name, _ in self.sub_data_containers]

    def __contains__(self, item):
        """
        return true if sub container name is in the sub data containers.

        :return: contains
        :rtype: bool
        """
        if isinstance(item, str):
            contains = False
            for name, data_container in self.sub_data_containers:
                if name == item:
                    contains = True
            return contains
        else:
            raise NotImplementedError('DataContainer.__contains__ not implemented for this type: {}'.format(type(item)))

    def __getitem__(self, item: Union[str, int]):
        """
        If item is a string, then get then sub container with the same name as item in the list of sub data containers.
        If item is an int, then return a tuple of (current id, data input, expected output) for the given item index.

        :param item: sub data container str, or item index as int
        :type item: Union[str, int]
        :return: data container, or tuple of current ids, data inputs, expected outputs.
        :rtype: Union[DataContainer, Tuple]
        """
        if isinstance(item, str):
            for name, data_container in self.sub_data_containers:
                if name == item:
                    return data_container
            raise KeyError("sub_data_container {} not found in data container".format(item))
        else:
            if self.current_ids is None:
                current_ids = [None] * len(self)
            else:
                current_ids = self.current_ids[item]

            return current_ids, self.data_inputs[item], self.expected_outputs[item]

    def tolist(self):
        current_ids = self.current_ids
        data_inputs = self.data_inputs
        expected_outputs = self.expected_outputs

        if isinstance(self.current_ids, np.ndarray):
            current_ids = self.current_ids.tolist()

        if isinstance(self.data_inputs, np.ndarray):
            data_inputs = self.data_inputs.tolist()

        if isinstance(self.expected_outputs, np.ndarray):
            expected_outputs = self.expected_outputs.tolist()

        self.set_current_ids(current_ids)
        self.set_data_inputs(data_inputs)
        self.set_expected_outputs(expected_outputs)

        return self

    def tolistshallow(self):
        self.set_current_ids(list(self.current_ids))
        self.set_data_inputs(list(self.data_inputs))
        self.set_expected_outputs(list(self.expected_outputs))

        return self

    def to_numpy(self):
        self.set_current_ids(np.array(self.current_ids))
        self.set_data_inputs(np.array(self.data_inputs))
        self.set_expected_outputs(np.array(self.expected_outputs))
        return self

    def __iter__(self):
        """
        Iter method returns a zip of all of the current_ids, data_inputs, and expected_outputs in the data container.

        :return: iterator of tuples containing current_ids, data_inputs, and expected outputs
        :rtype: Iterator[Tuple]
        """
        current_ids = self.current_ids
        if self.current_ids is None:
            current_ids = [None] * len(self.data_inputs)

        expected_outputs = self.expected_outputs
        if self.expected_outputs is None:
            expected_outputs = [None] * len(self.data_inputs)

        return zip(current_ids, self.data_inputs, expected_outputs)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.__class__.__name__ + "(current_ids=" + repr(list(self.current_ids)) + ", summary_id=" + repr(
            self.summary_id)

    def __len__(self):
        return len(self.data_inputs)


class ExpandedDataContainer(DataContainer):
    """
    Sub class of DataContainer to expand data container dimension.

    .. seealso::
        :class:`DataContainer`,
    """

    def __init__(self, data_inputs, current_ids, expected_outputs, summary_id, old_current_ids):
        DataContainer.__init__(
            self,
            data_inputs=data_inputs,
            current_ids=current_ids,
            summary_id=summary_id,
            expected_outputs=expected_outputs,
        )

        self.old_current_ids = old_current_ids

    def reduce_dim(self) -> 'DataContainer':
        """
        Reduce DataContainer to its original shape with a list of multiple current_ids, data_inputs, and expected outputs.

        :return: reduced data container
        :rtype: DataContainer
        """
        if len(self.data_inputs) != 1:
            raise ValueError(
                'Invalid Expanded Data Container. Please create ExpandedDataContainer with ExpandedDataContainer.create_from(data_container) method.')

        return DataContainer(
            data_inputs=self.data_inputs[0],
            current_ids=self.old_current_ids,
            summary_id=self.summary_id,
            expected_outputs=self.expected_outputs[0],
            sub_data_containers=self.sub_data_containers
        )

    @staticmethod
    def create_from(data_container: DataContainer) -> 'ExpandedDataContainer':
        """
        Create ExpandedDataContainer with the given summary hash for the single current id.

        :param data_container: data container to transform
        :type data_container: DataContainer
        :return: expanded data container
        :rtype: ExpandedDataContainer
        """
        return ExpandedDataContainer(
            data_inputs=[data_container.data_inputs],
            current_ids=[data_container.summary_id],
            summary_id=data_container.summary_id,
            expected_outputs=[data_container.expected_outputs],
            old_current_ids=data_container.current_ids
        )


class ZipDataContainer(DataContainer):
    """
    Sub class of DataContainer to zip two data sources together.

    .. seealso::
        :class:`DataContainer`,
    """

    @staticmethod
    def create_from(data_container: DataContainer, *other_data_containers: DataContainer) -> 'ZipDataContainer':
        """
        Create ZipDataContainer that merges two data sources together.

        :param data_container: data container to transform
        :type data_container: DataContainer
        :param other_data_containers: other data containers to zip with data container
        :type other_data_containers: List[DataContainer]
        :return: expanded data container
        :rtype: ExpandedDataContainer
        """
        new_data_inputs = []

        for i, (_, di, eo) in enumerate(data_container):
            new_data_input = [di]

            for other_data_container in other_data_containers:
                _, di, eo = other_data_container[i]
                new_data_input.append(di)

            new_data_inputs.append(tuple(new_data_input))

        return ZipDataContainer(
            data_inputs=new_data_inputs,
            current_ids=data_container.current_ids,
            summary_id=data_container.summary_id,
            expected_outputs=data_container.expected_outputs,
            sub_data_containers=data_container.sub_data_containers
        )

    def concatenate_inner_features(self):
        """
        Concatenate inner features from zipped data inputs.
        Broadcast data inputs if the dimension is smaller.
        """
        new_data_inputs = [di[0] for di in self.data_inputs]

        for i, data_input in enumerate(self.data_inputs):
            for di in data_input[1:]:
                new_data_inputs[i] = _inner_concatenate_np_array(new_data_inputs[i], di)

        self.set_data_inputs(new_data_inputs)


class ListDataContainer(DataContainer):
    """
    Sub class of DataContainer to perform list operations.
    It allows to perform append, and concat operations on a DataContainer.

    .. seealso::
        :class:`DataContainer`
    """

    def __init__(self, data_inputs: Any, current_ids=None, summary_id=None, expected_outputs: Any = None, sub_data_containers=None):
        DataContainer.__init__(self, data_inputs, current_ids, summary_id, expected_outputs, sub_data_containers)
        self.tolistshallow()

    @staticmethod
    def empty(original_data_container: DataContainer = None) -> 'ListDataContainer':
        if original_data_container is None:
            sub_data_containers = []
        else:
            sub_data_containers = original_data_container.sub_data_containers

        return ListDataContainer([], [], [], sub_data_containers=sub_data_containers)

    def append(self, current_id: str, data_input: Any, expected_output: Any):
        """
        Append a new data input to the DataContainer.

        :param current_id: current id for the data input
        :type current_id: str
        :param data_input: data input
        :param expected_output: expected output
        :return:
        """
        self.current_ids.append(current_id)
        self.data_inputs.append(data_input)
        self.expected_outputs.append(expected_output)

    def append_data_container_in_data_inputs(self, other: DataContainer) -> 'ListDataContainer':
        """
        Append a data container to the data inputs of this data container.

        :param other: data container
        :type other: DataContainer
        :return:
        """
        self.data_inputs.append(other)
        return self

    def append_data_container(self, other: DataContainer) -> 'ListDataContainer':
        """
        Append a data container to the DataContainer.

        :param other: data container
        :type other: DataContainer
        :return:
        """
        self.current_ids.append(other.current_ids)
        self.data_inputs.append(other.data_inputs)
        self.expected_outputs.append(other.expected_outputs)

        return self

    def concat(self, data_container: DataContainer):
        """
        Concat the given data container to the current data container.

        :param data_container: data container
        :type data_container: DataContainer
        :return:
        """
        data_container.tolistshallow()

        self.current_ids.extend(data_container.current_ids)
        self.data_inputs.extend(data_container.data_inputs)
        self.expected_outputs.extend(data_container.expected_outputs)

        return self


def _inner_concatenate_np_array(np_array, np_array_to_zip):
    """
    Concatenate numpy arrays on the last axis using expand dim, and broadcasting.

    :param np_array: data container
    :type np_array: np.ndarray
    :param np_array_to_zip: numpy array to zip with the other
    :type np_array_to_zip: np.ndarray
    :return: concatenated np array
    :rtype: np.ndarray
    """
    while len(np_array_to_zip.shape) < len(np_array.shape):
        np_array_to_zip = np.expand_dims(np_array_to_zip, axis=-1)

    target_shape = tuple(list(np_array.shape[:-1]) + [np_array_to_zip.shape[-1]])
    np_array_to_zip = np.broadcast_to(np_array_to_zip, target_shape)

    return np.concatenate((np_array, np_array_to_zip), axis=-1)



"""
Streaming Parallel Data Processing
===================================================================

Neuraxle steps for streaming data in parallel in the pipeline

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import warnings
from abc import abstractmethod
from multiprocessing import Queue
from multiprocessing.context import Process
from threading import Thread
from typing import Tuple, List, Union, Iterable

from neuraxle.base import NamedTupleList, ExecutionContext, BaseStep, MetaStep, BaseSaver, _FittableStep, \
    BaseTransformer, NonFittableMixin
from neuraxle.data_container import DataContainer, ListDataContainer
from neuraxle.pipeline import Pipeline, MiniBatchSequentialPipeline, Joiner
from neuraxle.steps.numpy import NumpyConcatenateOuterBatch


class ObservableQueueMixin:
    """
    A class to represent a step that can put items in a queue.
    It can also notify other queues that have subscribed to him using subscribe.

    .. seealso::
        :class:`BaseStep`,
        :class:`QueuedPipelineTask`,
        :class:`QueueWorker`,
        :class:`BaseQueuedPipeline`,
        :class:`ParallelQueuedPipeline`,
        :class:`SequentialQueuedPipeline`
    """

    def __init__(self, queue):
        self.queue = queue
        self.observers = []
        self._add_observable_queue_step_saver()

    def teardown(self):
        self.queue = None
        return self

    def _add_observable_queue_step_saver(self):
        if not hasattr(self, 'savers'):
            warnings.warn(
                'Please initialize Mixins in the good order. ObservableQueueMixin should be initialized after '
                'Appending the ObservableQueueStepSaver to the savers. Saving might fail.'
            )
            self.savers = [ObservableQueueStepSaver()]
        else:
            self.savers.append(ObservableQueueStepSaver())

    def subscribe(self, observer_queue_worker: 'ObservableQueueMixin') -> 'ObservableQueueMixin':
        """
        Subscribe a queue worker.
        The subscribed queue workers get notified when :func:`~neuraxle.distributed.streaming.ObservableQueueMixin.notify` is called.
        """
        self.observers.append(observer_queue_worker.queue)
        return self

    def get(self) -> 'QueuedPipelineTask':
        """
        Get last item in queue.
        """
        return self.queue.get()

    def put(self, value: DataContainer):
        """
        Put a queued pipeline task in queue.
        """
        self.queue.put(QueuedPipelineTask(step_name=self.name, data_container=value.copy()))

    def notify(self, value):
        """
        Notify all subscribed queue workers
        """
        for observer in self.observers:
            observer.put(value)


class QueuedPipelineTask(object):
    """
    Data object to contain the tasks processed by the queued pipeline.

    .. seealso::
        :class:`QueueWorker`,
        :class:`BaseQueuedPipeline`,
        :class:`ParallelQueuedPipeline`,
        :class:`SequentialQueuedPipeline`
    """

    def __init__(self, data_container, step_name=None):
        self.step_name = step_name
        self.data_container = data_container


class ObservableQueueStepSaver(BaseSaver):
    """
    Saver for observable queue steps.

    .. seealso::
        :class:`QueueWorker`,
        :class:`neuraxle.base.BaseSaver`,
        :class:`BaseQueuedPipeline`,
        :class:`ParallelQueuedPipeline`,
        :class:`SequentialQueuedPipeline`
    """

    def save_step(self, step: BaseTransformer, context: 'ExecutionContext') -> BaseTransformer:
        step.queue = None
        step.observers = []
        return step

    def can_load(self, step: BaseTransformer, context: 'ExecutionContext') -> bool:
        return True

    def load_step(self, step: 'BaseTransformer', context: 'ExecutionContext') -> 'BaseTransformer':
        step.queue = Queue()
        return step


class QueueWorker(ObservableQueueMixin, MetaStep):
    """
    Start multiple Process or Thread that process items from the queue of batches to process.
    It is both an observable, and observer.
    It notifies the results of the wrapped step handle transform method.
    It receives the next data container to process.

    .. seealso::
        :class:`Observer`,
        :class:`Observable`,
        :class:`MetaStepMixin`,
        :class:`BaseStep`
    """

    def __init__(
            self,
            wrapped: BaseTransformer,
            max_queue_size: int,
            n_workers: int,
            use_threading: bool,
            additional_worker_arguments=None,
            use_savers=False
    ):
        if not additional_worker_arguments:
            additional_worker_arguments = [[] for _ in range(n_workers)]

        MetaStep.__init__(self, wrapped)
        ObservableQueueMixin.__init__(self, Queue(maxsize=max_queue_size))

        self.use_threading: bool = use_threading
        self.workers: List[Process] = []
        self.n_workers: int = n_workers
        self.observers: List[Queue] = []
        self.additional_worker_arguments = additional_worker_arguments
        self.use_savers = use_savers

    def start(self, context: ExecutionContext):
        """
        Start multiple processes or threads with the worker function as a target.

        :param context: execution context
        :type context: ExecutionContext
        :return:
        """
        target_function = worker_function
        if self.use_savers:
            self.save(context, full_dump=True)
            target_function = worker_function

        self.workers = []
        for _, worker_arguments in zip(range(self.n_workers), self.additional_worker_arguments):
            if self.use_threading:
                p = Thread(target=target_function, args=(self, context,  self.use_savers, worker_arguments))
            else:
                p = Process(target=target_function, args=(self, context, self.use_savers, worker_arguments))

            p.daemon = True
            p.start()
            self.workers.append(p)

    def teardown(self):
        """
        Stop all processes on teardown.

        :return: teardowned self
        """
        self.stop()
        return self

    def stop(self):
        """
        Stop all of the workers.

        :return:
        """
        if not self.use_threading:
            [w.terminate() for w in self.workers]

        self.workers = []
        self.observers = []


def worker_function(queue_worker: QueueWorker, context: ExecutionContext, use_savers: bool, additional_worker_arguments):
    """
    Worker function that transforms the items inside the queue of items to process.

    :param queue_worker: step to transform
    :param context: execution context
    :param use_savers: use savers
    :param additional_worker_arguments: any additional arguments that need to be passed to the workers
    :return:
    """
    step = queue_worker.get_step()
    if use_savers:
        saved_queue_worker: QueueWorker = context.load(queue_worker.get_name())
        step = saved_queue_worker.get_step()

    additional_worker_arguments = tuple(
        additional_worker_arguments[i: i + 2] for i in range(0, len(additional_worker_arguments), 2)
    )

    for argument_name, argument_value in additional_worker_arguments:
        step.__dict__.update({argument_name: argument_value})

    while True:
        task: QueuedPipelineTask = queue_worker.get()
        summary_id = task.data_container.summary_id
        data_container = step.handle_transform(task.data_container, context)
        data_container = data_container.set_summary_id(summary_id)
        queue_worker.notify(QueuedPipelineTask(step_name=queue_worker.name, data_container=data_container))


QueuedPipelineStepsTuple = Union[
    BaseTransformer,  # step
    Tuple[int, BaseTransformer],  # (n_workers, step)
    Tuple[str, BaseTransformer],  # (step_name, step)
    Tuple[str, int, BaseTransformer],  # (step_name, n_workers, step)
    Tuple[str, int, int, BaseTransformer],  # (step_name, n_workers, max_queue_size, step)
    Tuple[str, int, List[Tuple], BaseTransformer],  # (step_name, n_workers, additional_worker_arguments, step)
    Tuple[str, int, List[Tuple], BaseTransformer]  # (step_name, n_workers, additional_worker_arguments, step)
]


class BaseQueuedPipeline(MiniBatchSequentialPipeline):
    """
    Sub class of :class:`Pipeline`.
    Transform data in many pipeline steps at once in parallel in the pipeline using multiprocessing Queues.

    Example usage :

    .. code-block:: python

        # step name, step

        p = QueuedPipeline([
            ('step_a', Identity()),
            ('step_b', Identity()),
        ], n_workers=1, batch_size=10, max_queue_size=10)

        # step name, number of workers, step

        p = QueuedPipeline([
            ('step_a', 1, Identity()),
            ('step_b', 1, Identity()),
        ], batch_size=10, max_queue_size=10)

        # step name, number of workers, and max size

        p = QueuedPipeline([
            ('step_a', 1, 10, Identity()),
            ('step_b', 1, 10, Identity()),
        ], batch_size=10)

        # step name, number of workers for each step, and additional argument for each worker

        p = QueuedPipeline([
            ('step_a', 1, [('host', 'host1'), ('host', 'host2')], 10, Identity())
        ], batch_size=10)

        # step name, number of workers for each step, additional argument for each worker, and max size

        p = QueuedPipeline([
            ('step_a', 1, [('host', 'host1'), ('host', 'host2')], 10, Identity())
        ], batch_size=10)


    .. seealso::
        :class:`QueueWorker`,
        :class:`QueueJoiner`,
        :class:`CustomPipelineMixin`,
        :class:`Pipeline`
    """

    def __init__(
            self,
            steps: List[QueuedPipelineStepsTuple],
            batch_size,
            n_workers_per_step=None,
            max_queue_size=None,
            data_joiner=None,
            use_threading=False,
            use_savers=False,
            cache_folder=None
    ):
        if data_joiner is None:
            data_joiner = NumpyConcatenateOuterBatch()
        self.data_joiner = data_joiner
        self.max_queue_size = max_queue_size
        self.batch_size = batch_size
        self.n_workers_per_step = n_workers_per_step
        self.use_threading = use_threading
        self.use_savers = use_savers

        MiniBatchSequentialPipeline.__init__(self, steps=self._initialize_steps_as_tuple(steps),
                                             cache_folder=cache_folder)
        self._refresh_steps()

    def _initialize_steps_as_tuple(self, steps):
        """
        Wrap each step by a :class:`QueueWorker` to  allow data to flow in many pipeline steps at once in parallel.

        :param steps: (name, n_workers, step)
        :type steps: NameNWorkerStepTupleList
        :return: steps as tuple
        :rtype: NamedTupleList
        """
        steps_as_tuple: NamedTupleList = []
        for step in steps:
            queue_worker = self._create_queue_worker(step)
            steps_as_tuple.append((queue_worker.name, queue_worker))

        steps_as_tuple.append(('queue_joiner', QueueJoiner(batch_size=self.batch_size)))

        return steps_as_tuple

    def _create_queue_worker(self, step: QueuedPipelineStepsTuple):
        name, n_workers, additional_worker_arguments, max_queue_size, actual_step = self._get_step_params(step)

        return QueueWorker(
            actual_step,
            n_workers=n_workers,
            use_threading=self.use_threading,
            max_queue_size=max_queue_size,
            additional_worker_arguments=additional_worker_arguments,
            use_savers=self.use_savers
        ).set_name('QueueWorker{}'.format(name))

    def _get_step_params(self, step):
        """
        Return all params necessary to create the QueuedPipeline for the given step.

        :param step: tuple
        :type step: QueuedPipelineStepsTupleList

        :return: return name, n_workers, max_queue_size, actual_step
        :rtype: tuple(str, int, int, BaseStep)
        """
        if isinstance(step, BaseTransformer):
            actual_step = step
            name = step.name
            max_queue_size = self.max_queue_size
            n_workers = self.n_workers_per_step
            additional_arguments = []
        elif len(step) == 2:
            if isinstance(step[0], str):
                name, actual_step = step
                n_workers = self.n_workers_per_step
            else:
                n_workers, actual_step = step
                name = actual_step.name
            max_queue_size = self.max_queue_size
            additional_arguments = []
        elif len(step) == 3:
            name, n_workers, actual_step = step
            max_queue_size = self.max_queue_size
            additional_arguments = []
        elif len(step) == 4:
            if isinstance(step[2], Iterable):
                name, n_workers, additional_arguments, actual_step = step
                max_queue_size = self.max_queue_size
            else:
                name, n_workers, max_queue_size, actual_step = step
                additional_arguments = []
        elif len(step) == 5:
            name, n_workers, additional_arguments, max_queue_size, actual_step = step
        else:
            raise Exception('Invalid Queued Pipeline Steps Shape.')

        return name, n_workers, additional_arguments, max_queue_size, actual_step

    def _will_process(self, data_container: DataContainer, context: ExecutionContext) -> (DataContainer, ExecutionContext):
        """
        Setup streaming pipeline before any handler methods.

        :param data_container: data container
        :param context: execution context
        :return:
        """
        self.setup(context=context)
        return data_container, context

    def setup(self, context: ExecutionContext = None) -> 'BaseTransformer':
        """
        Connect the queued workers together so that the data can correctly flow through the pipeline.

        :param context: execution context
        :return: step
        :rtype: BaseStep
        """
        if not self.is_initialized:
            self.connect_queued_pipeline()
        super().setup(context=context)
        return self

    def fit_transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> ('Pipeline', DataContainer):
        """
        Fit transform sequentially if any step is fittable. Otherwise transform in parallel.

        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return:
        """
        all_steps_are_not_fittable = True

        for _, step in self[:-1]:
            if isinstance(step.get_step(), _FittableStep) and not isinstance(step.get_step(), NonFittableMixin):
                all_steps_are_not_fittable = False

        if all_steps_are_not_fittable:
            data_container = self.transform_data_container(data_container, context)
            data_container = self._did_transform(data_container, context)
            return self, data_container

        self.is_invalidated = True

        return super().fit_transform_data_container(data_container, context)

    def transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        """
        Transform data container

        :param data_container: data container to transform.
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: data container
        """
        data_container_batches = data_container.convolved_1d(stride=self.batch_size, kernel_size=self.batch_size)

        n_batches = self.get_n_batches(data_container)
        self[-1].set_n_batches(n_batches)

        for name, step in self[:-1]:
            step.start(context)

        batch_index = 0
        for data_container_batch in data_container_batches:
            self.send_batch_to_queued_pipeline(batch_index=batch_index, data_container=data_container_batch)
            batch_index += 1

        data_container = self[-1].join(original_data_container=data_container)
        return data_container

    def _did_transform(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        """
        Stop all of the workers after transform. Also, join the data using self.data_joiner.

        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: data container
        :rtype: DataContainer
        """
        for name, step in self[:-1]:
            step.stop()

        return self.data_joiner.handle_transform(data_container, context)

    @abstractmethod
    def get_n_batches(self, data_container) -> int:
        """
        Get the total number of batches that the queue joiner is supposed to receive.

        :param data_container: data container to transform
        :type data_container: DataContainer
        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def connect_queued_pipeline(self):
        """
        Connect all the queued workers together so that the data can flow through each step.

        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def send_batch_to_queued_pipeline(self, batch_index: int, data_container: DataContainer):
        """
        Send batches to queued pipeline. It is blocking if there is no more space available in the multiprocessing queues.
        Workers might return batches in a different order, but the queue joiner will reorder them at the end.
        The queue joiner will use the summary ids to reorder all of the received batches.

        :param batch_index: batch index
        :param data_container: data container batch
        :return:
        """
        raise NotImplementedError()


class SequentialQueuedPipeline(BaseQueuedPipeline):
    """
    Using :class:`QueueWorker`, run all steps sequentially even if they are in separate processes or threads.

    .. seealso::
        :class:`QueueWorker`,
        :class:`BaseQueuedPipeline`,
        :class:`ParallelQueuedPipeline`,
        :class:`QueueJoiner`,
        :class:`Observer`,
        :class:`Observable`
    """

    def get_n_batches(self, data_container) -> int:
        """
        Get the number of batches to process.

        :param data_container: data container to transform
        :return: number of batches
        """
        return data_container.get_n_batches(self.batch_size)

    def connect_queued_pipeline(self):
        """
        Sequentially connect of the queued workers.

        :return:
        """
        for i, (name, step) in enumerate(self[1:]):
            self[i].subscribe(step)

    def send_batch_to_queued_pipeline(self, batch_index: int, data_container: DataContainer):
        """
        Send batches to process to the first queued worker.

        :param batch_index: batch index
        :param data_container: data container batch
        :return:
        """
        data_container = data_container.set_summary_id(data_container.hash_summary())
        self[-1].summary_ids.append(data_container.summary_id)
        self[0].put(data_container)


class ParallelQueuedFeatureUnion(BaseQueuedPipeline):
    """
    Using :class:`QueueWorker`, run all steps in parallel using QueueWorkers.

    .. seealso::
        :class:`QueueWorker`,
        :class:`BaseQueuedPipeline`,
        :class:`SequentialQueuedPipeline`,
        :class:`QueueJoiner`,
        :class:`Observer`,
        :class:`Observable`
    """

    def get_n_batches(self, data_container):
        """
        Get the number of batches to process by the queue joiner.

        :return:
        """
        return data_container.get_n_batches(self.batch_size) * (len(self) - 1)

    def connect_queued_pipeline(self):
        """
        Connect the queue joiner to all of the queued workers to process data in parallel.

        :return:
        """
        for name, step in self[:-1]:
            step.subscribe(self[-1])

    def send_batch_to_queued_pipeline(self, batch_index: int, data_container: DataContainer):
        """
        Send batches to process to all of the queued workers.

        :param batch_index: batch index
        :param data_container: data container batch
        :return:
        """
        for name, step in self[:-1]:
            data_container = data_container.set_summary_id(data_container.hash_summary())
            self[-1].summary_ids.append(data_container.summary_id)
            step.put(data_container)


class QueueJoiner(ObservableQueueMixin, Joiner):
    """
    Observe the results of the queue worker of type :class:`QueueWorker`.
    Synchronize all of the workers together.

    .. seealso::
        :class:`QueuedPipeline`,
        :class:`Observer`,
        :class:`ListDataContainer`,
        :class:`DataContainer`
    """

    def __init__(self, batch_size, n_batches=None):
        self.n_batches_left_to_do = n_batches
        self.summary_ids = []
        self.result = {}
        Joiner.__init__(self, batch_size=batch_size)
        ObservableQueueMixin.__init__(self, Queue())

    def teardown(self) -> 'BaseTransformer':
        """
        Properly clean queue, summary ids, and results during teardown.

        :return: teardowned self
        """
        ObservableQueueMixin.teardown(self)
        Joiner.teardown(self)
        self.summary_ids = []
        self.result = {}
        return self

    def set_n_batches(self, n_batches):
        self.n_batches_left_to_do = n_batches

    def join(self, original_data_container: DataContainer) -> DataContainer:
        """
        Return the accumulated results received by the on next method of this observer.

        :return: transformed data container
        :rtype: DataContainer
        """
        while self.n_batches_left_to_do > 0:
            task: QueuedPipelineTask = self.queue.get()
            self.n_batches_left_to_do -= 1
            step_name = task.step_name

            if step_name not in self.result:
                self.result[step_name] = ListDataContainer(
                    current_ids=[],
                    data_inputs=[],
                    expected_outputs=[],
                    summary_id=task.data_container.summary_id
                )

            self.result[step_name].append_data_container_in_data_inputs(task.data_container)

        data_containers = self._join_all_step_results()
        self.result = {}
        return original_data_container.set_data_inputs(data_containers)

    def _join_all_step_results(self):
        """
        Concatenate all resulting data containers together.

        :return:
        """
        results = []
        for step_name, data_containers in self.result.items():
            step_results = self._join_step_results(data_containers)
            results.append(step_results)

        return results

    def _join_step_results(self, data_containers):
        # reorder results by summary id
        data_containers.data_inputs.sort(key=lambda dc: self.summary_ids.index(dc.summary_id))

        step_results = ListDataContainer.empty()
        for data_container in data_containers.data_inputs:
            data_container = data_container.set_summary_id(data_containers.data_inputs[-1].summary_id)
            step_results.concat(data_container)

        return step_results



"""
Neuraxle's Observable Classes
===================================================================
Base observable classes

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
from abc import abstractmethod
from typing import TypeVar, Generic

T = TypeVar('T')


class _Observable(Generic[T]):
    def __init__(self):
        self._observers = set()

    def subscribe(self, observer: '_Observer[T]'):
        self._observers.add(observer)

    def unsubscribe(self, observer: '_Observer[T]'):
        self._observers.discard(observer)

    def on_next(self, value: T):
        for observer in self._observers:
            observer.on_next(value)

    def on_complete(self, value: T):
        for observer in self._observers:
            observer.on_complete(value)

    def on_error(self, value: Exception):
        for observer in self._observers:
            observer.on_error(value)


class _Observer(Generic[T]):
    @abstractmethod
    def on_next(self, value: T):
        pass

    def on_complete(self, value: T):
        pass

    def on_error(self, value: Exception):
        pass



"""
Neuraxle's Column Transformer Steps
====================================

Pipeline steps to apply N-Dimensional column transformations to different columns.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
from typing import List, Tuple, Union

import numpy as np

from neuraxle.base import BaseStep, MetaStep, BaseTransformer
from neuraxle.pipeline import Pipeline
from neuraxle.steps.loop import ForEachDataInput
from neuraxle.union import FeatureUnion

ColumnSelectionType = Union[Tuple[int, BaseTransformer], Tuple[List[int], BaseTransformer], Tuple[slice, BaseTransformer]]
ColumnChooserTupleList = List[ColumnSelectionType]


class ColumnSelector2D(BaseTransformer):
    """
    A ColumnSelector2D selects column in a sequence.
    """

    def __init__(self, columns_selection: ColumnSelectionType):
        super().__init__()
        self.column_selection = columns_selection

    def transform(self, data_inputs):
        if isinstance(self.column_selection, range):
            self.column_selection = slice(
                self.column_selection.start,
                self.column_selection.stop,
                self.column_selection.step
            )

        if isinstance(self.column_selection, int):
            return np.expand_dims(np.array(data_inputs)[:, self.column_selection], axis=-1)

        if isinstance(self.column_selection, slice):
            return np.array(data_inputs)[:, self.column_selection]

        if isinstance(self.column_selection, list):
            columns = [
                np.expand_dims(np.array(data_inputs)[:, i], axis=-1)
                for i in self.column_selection
            ]
            return np.concatenate(columns, axis=-1)

        if self.column_selection is None:
            return data_inputs

        raise ValueError(
            'column selection type not supported : {0}\nSupported types'.format(
                self.column_selection,
                repr(ColumnSelectionType)
            ))


class ColumnsSelectorND(MetaStep):
    """
    ColumnSelectorND wraps a ColumnSelector2D by as many ForEachDataInput step
    as needed to select the last dimension.
    """

    def __init__(self, columns_selection, n_dimension=3):

        col_selector: ColumnSelector2D = ColumnSelector2D(columns_selection=columns_selection)
        for _ in range(min(0, n_dimension - 2)):
            col_selector = ForEachDataInput(col_selector)

        MetaStep.__init__(self, col_selector)
        self.n_dimension = n_dimension


class ColumnTransformer(FeatureUnion):
    """
    A ColumnChooser can apply custom transformations to different columns.
    The ColumnChooser accepts a list of tuples for the transformations,
    and will name the steps accordingly (because of the TruncableSteps' constructor)
    by converting each indexer object to a string. Indexer objects can be ranges, an int, or a list of ints.
    The input data can be `N`-dimensionnal (ND), in which case the axis must be specified. The columns
    data passed to the sub-steps will still be ND.

    Usage example:

    .. code-block:: python

        ColumnChooser([
            (range(0, 2), CyclicTimes()),
            (3, CategoricalEnum(categories_count=5, starts_at_zero=True)),
            (4, CategoricalEnum(categories_count=5, starts_at_zero=True)),
            ([10, 13, 15], CategoricalEnum(categories_count=5, starts_at_zero=True)),
        ])

    .. seealso::
        :class:`~neuraxle.union.FeatureUnion`,
    """

    def __init__(self, column_chooser_steps_as_tuple: ColumnChooserTupleList, n_dimension: int = 3):
        # Make unique names from the indices in case we have many steps for transforming the same column(s).
        self.string_indices = [
            str(name) + "_" + str(step.__class__.__name__)
            for name, step in column_chooser_steps_as_tuple
        ]

        super().__init__([
            (string_indices, Pipeline([
                ColumnsSelectorND(indices, n_dimension=n_dimension),
                step
            ]))
            for string_indices, (indices, step) in zip(self.string_indices, column_chooser_steps_as_tuple)
        ])
